#include "squidwall.h"
#include "strallocX.h"

/**
 * Copyright (C) 2005-2006 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

static int setup_acls(int user, int host, int global);
static void setup(void);

/* caching of content for learning: on/off via 1/0 */
int cache=0;

/* enable/disable global acl defaults (for unknown hosts/users) */
int globals=0; /* they are defined @ /etc/squidwall/control/! */

/* max size for content/virus scanning */
int maxsize=MAXSIZE;

/* what should be logged */
int logtodo=LOGTODO;

/* caching of results */
int memcache=MEMCACHE;

/* max hitpoints for phrase filtering */
int pfilter_max=PFILTER_MAX;

/* default hitpoints, if configuration is buggy */
int pfilter_default=PFILTER_DEFAULT;

/* default timeout, for sqlite3 busy requests */
int sqlite_timeout=SQLITE_TIMEOUT;

/* pointer to the infos from squid @ fd=0 */
char *squid_url;
char *squid_ip;
char *squid_host;
char *squid_user;
char *squid_method;
unsigned squid_nr;

/* lists */
LIST_HEAD(hosts);
LIST_HEAD(users);
LIST_HEAD(redirects);
LIST_HEAD(bfilter);
LIST_HEAD(bcfilter);
LIST_HEAD(brfilter);
LIST_HEAD(cfilter);
LIST_HEAD(crfilter);
LIST_HEAD(pfilter);
LIST_HEAD(ufilter);
LIST_HEAD(ufilter_md5);
LIST_HEAD(ufilter_sha1);
LIST_HEAD(urfilter);
LIST_HEAD(wfilter);
LIST_HEAD(wrfilter);

/* defined filters of current ip / host */
stralloc *bfilter_host;
stralloc *bcfilter_host;
stralloc *brfilter_host;
stralloc *cfilter_host;
stralloc *crfilter_host;
stralloc *pfilter_host;
stralloc *ufilter_host;
stralloc *ufilter_md5_host;
stralloc *ufilter_sha1_host;
stralloc *urfilter_host;
stralloc *wfilter_host;
stralloc *wrfilter_host;

/* defined filters of current user */
stralloc *bfilter_user;
stralloc *bcfilter_user;
stralloc *brfilter_user;
stralloc *cfilter_user;
stralloc *crfilter_user;
stralloc *pfilter_user;
stralloc *ufilter_user;
stralloc *ufilter_md5_user;
stralloc *ufilter_sha1_user;
stralloc *urfilter_user;
stralloc *wfilter_user;
stralloc *wrfilter_user;

/* defined global filters */
stralloc *bfilter_globals;
stralloc *bcfilter_globals;
stralloc *brfilter_globals;
stralloc *cfilter_globals;
stralloc *crfilter_globals;
stralloc *pfilter_globals;
stralloc *ufilter_globals;
stralloc *ufilter_md5_globals;
stralloc *ufilter_sha1_globals;
stralloc *urfilter_globals;
stralloc *wfilter_globals;
stralloc *wrfilter_globals;

/* destination websites for squid */
char *redir_badurl=REDIR_BADURL;
char *redir_badcont=REDIR_BADCONT;
char *redir_baduser=REDIR_BADUSER;
char *redir_badhost=REDIR_BADHOST;
char *redir_safesite=REDIR_SAFESITE;
char *redir_virusfound=REDIR_VIRUSFOUND;
char *redir_bannerfound=REDIR_BANNERFOUND;
char *redir_webinterface=REDIR_WEBINTERFACE;

/* misc directories */
char *cache_dir=CACHE_DIR;
char *host_dir=HOST_DIR;
char *user_dir=USER_DIR;
char *redir_dir=REDIR_DIR;
char *global_dir=GLOBAL_DIR;
char *bfilter_dir=BFILTER_DIR;
char *bcfilter_dir=BCFILTER_DIR;
char *brfilter_dir=BRFILTER_DIR;
char *cfilter_dir=CFILTER_DIR;
char *crfilter_dir=CRFILTER_DIR;
char *pfilter_dir=PFILTER_DIR;
char *ufilter_dir=UFILTER_DIR;
char *ufilter_md5_dir=UFILTER_MD5_DIR;
char *ufilter_sha1_dir=UFILTER_SHA1_DIR;
char *urfilter_dir=URFILTER_DIR;
char *wfilter_dir=WFILTER_DIR;
char *wrfilter_dir=WRFILTER_DIR;

/* ip's and ports */
struct ip_address clamd_ip;
struct ip_address proxy_ip;

unsigned clamd_port;
unsigned clams_port;
unsigned proxy_port;

char *proxy=PROXY;
char *logfile=LOGFILE;
char *clamav_tcp=CLAMAV_TCP;
char *clamav_socket=CLAMAV_SOCKET;

/* decoded infos */
stralloc ip;
stralloc url;
stralloc host;
stralloc user;
stralloc method;

/* extracted infos (e.g. http://www.example.org/some/path/) */
stralloc url_host; /* -> www.example.org */
stralloc url_path; /* -> /some/path/ */

/* the final destination site */
stralloc redirsite;

/* infors for the redirect site */
stralloc clamav;
stralloc filter;
stralloc virus;

/**
 * setup_acls - setup the acl's according to some definitions
 *
 * @param user: current user acl's (defined @ /etc/squidwall/users/!)
 * @param host: current host/ip acl's (defined @ /etc/squidwall/hosts/!)
 * @param global: current global acl's (defined @ /etc/squidwall/globals/!)
 *
 * @return: ACL_ALLOW  -> the user/host can pass all filters
 *          ACL_FILTER -> filter have to be used
 *          ACL_DENY   -> the user/host will be denied
 */
static int setup_acls(int user, int host, int global)
{
	/* user or host get denied */
	if (user == ACL_DENY) {
		redir_site(REDIR_TYPE_USER);
		return ACL_DENY;
	}

	if (host == ACL_DENY) {
		redir_site(REDIR_TYPE_HOST);
		return ACL_DENY;
	}

	/* user or host get filtered */
	if (user == ACL_FILTER) return ACL_FILTER;
	if (host == ACL_FILTER) return ACL_FILTER;

	/* user and host are allowed or undefined, no globals are used */
	if ((user|host) == ACL_ALLOW) {
		return ACL_ALLOW;
	}

	if (global == ACL_DENY) {
		redir_site(REDIR_TYPE_HOST);
		return ACL_DENY;
	}

	/* return the global settings */
	return global;
}

/* read configuration environment variables */
static void setup(void)
{
	/* init the stralloc's */
	stralloc_init(&ip);
	stralloc_init(&url);
	stralloc_init(&host);
	stralloc_init(&user);
	stralloc_init(&method);
	stralloc_init(&redirsite);

	stralloc_init(&clamav);
	stralloc_init(&filter);
	stralloc_init(&virus);

	stralloc_init(&url_host);
	stralloc_init(&url_host);

	signal(SIGPIPE,SIG_IGN);
	signal(SIGALRM,SIG_IGN);

	log_init(1);
#if WANT_SQLITE
	/* creates our tables, if needed and connects to the database */
	cache_init();
#endif
	memcache_init();
	read_acls();
	log_flush();

	return;
}

/* int main(int argc, char **argv) */
int main(void)
{
	unsigned int i;
	stralloc sa, out;
	int r, flags, todo, state;

	if (!io_fd(0)) return 111;
	if (!io_fd(1)) return 111;

	setup();
	stralloc_init(&sa);
	stralloc_init(&out);
	for (;;) {
		flags=ACL_NONE; todo=TODO_NONE; state=0;

		/* free old stuff */
		if (!stralloc_copys(&sa, "")) die_nomem();
		if (!stralloc_copys(&out, "")) die_nomem();

		if (!stralloc_copys(&ip, "")) die_nomem();
		if (!stralloc_copys(&url, "")) die_nomem();
		if (!stralloc_copys(&host, "")) die_nomem();
		if (!stralloc_copys(&user, "")) die_nomem();
		if (!stralloc_copys(&method, "")) die_nomem();
		if (!stralloc_copys(&redirsite, "")) die_nomem();

		for (;;) {
			io_wantread(0); io_wait(); io_dontwantread(0);
			r=buffer_getline_sa(buffer_0, &sa);
			if (r == 0) return 0; /* exiting ... */
			if (r == 1) break;
		}
		sa.s[sa.len-1]=0; /* remove the newline */

		/* format_v1: "url ip/host user method"
		 * format_v2: "nr url ip/host user method urlgroup"
		 *
		 * sample1: http://www.mcmilk.de/ 192.168.100.10/aspire mcmilk GET
		 * sample2: www.blabla.de:443 192.168.100.6/p28tr - CONNECT
		 *
		 * state=0 url=http://www.mcmilk.de/
		 * state=1 ip=192.168.100.10
		 * state=2 host=aspire
		 * state=3 user=root
		 * state=4 method=GET
		 */

		squid_url=sa.s;
		for (i=0; i<=sa.len; i++) {
			switch (state) {
			case 0:
			if (sa.s[i] == ' ') {
				sa.s[i]=0;state=1;
				/* look for url_rewrite_concurrency */
				r = scan_uint(sa.s, &squid_nr);
				if (r) {
					set(flags, ACL_MULTI);
					debug_ulong(DEBUG_SQUIDWALL, "threading: ",
						squid_nr);
					squid_url=sa.s+r+1;
					while (i++)
					if (sa.s[i] == ' ') {
						sa.s[i]=0; break;
					}
				}
				squid_ip=sa.s+i+1;
				scan_url();
				url_host.s[url_host.len]=0;
				url_path.s[url_path.len]=0;
				debug_s3(DEBUG_SQUIDWALL, "scan_url(): host(",
					url_host.s, ")");
				debug_s3(DEBUG_SQUIDWALL, "scan_url(): path(",
					url_path.s, ")");
			}
			break;

			case 1:
			if (sa.s[i] == '/') {
				sa.s[i]=0; state=2; squid_host=sa.s+i+1;
			}
			break;

			case 2:
			if (sa.s[i] == ' ') {
				sa.s[i]=0; state=3; squid_user=sa.s+i+1;
			}
			break;

			case 3:
			if (sa.s[i] == ' ') {
				sa.s[i]=0; state=4; squid_method=sa.s+i+1;
			}
			break;

			case 4:
			/* new, since 2.6-stable1 (url groups) */
			if (sa.s[i] == ' ') { sa.s[i]=0; state=5; }
			break;

			} /* switch() */
		}

		if (state != 4 && state != 5) {
			log_headline("error");
			log_s("incorrect input from squid!\n");
			log_flush();
			continue;
		}

		state=0;
		if (str_start(squid_method, "HEAD")) state=1;
		if (str_start(squid_method, "POST")) state=1;
		if (str_start(squid_method, "GET")) state=1;

		if (state != 1) {
			set(flags, ACL_REDIR);
			/* eg: redirecting CONNECT will cause failures */
		}

		/* squid gives us urlencoded things - we decode them first! */
		if (!scan_to_sa(scan_urlencoded,  squid_ip, &ip)) die_nomem();
		if (!scan_to_sa(scan_urlencoded2, squid_url, &url)) die_nomem();
		if (!scan_to_sa(scan_urlencoded,  squid_host, &host)) die_nomem();
		if (!scan_to_sa(scan_urlencoded2, squid_user, &user)) die_nomem();
		if (!scan_to_sa(scan_urlencoded,  squid_method, &method)) die_nomem();

		ip.s[ip.len]=0;
		url.s[url.len]=0;
		host.s[host.len]=0;
		user.s[user.len]=0;
		method.s[method.len]=0;

		/* we check in this order:
		 *
		 *
		 * 1  is the url http://squidwall -> redirect to webinterface
		 * 2  check for redirect sites -> redirect if needed
		 * 3. check the acl's (global, user, host, ip)
		 * 4. look for ad's / banners
		 * 5. look for whitelisted url's
		 * 6. look for blacklisted url's
		 * 7. check for bad content
		 *    - pfilter, cfilter, crfilter
		 *    - antivirus
		 *    - bcfilter (banner via content regex filter)
		 *    - learning / classifying
		 */

		/* the internet explorer will not redirect per default
		 * - it will just try to resolve the host 'squidwall' and then
		 *   redirect to your favorite search engine ...
		 * - just use http://squidwall.lala.de/
		 */
		if (str_start(squid_url, "http://squidwall")) {
			redir_site(REDIR_TYPE_CONFIG);
			set(flags, ACL_REDIR);
			goto out;
		}

		/* just pass through */
		if (str_start(squid_url, redir_safesite)) {
			goto out;
		}

		/* check, if the url matches some user defined redirect site */
		if (check_redirect()) {
			/* stralloc redirsite is set in the function */
			goto out;
		}

		/* recheck the filter definitions and acl's */
		read_acls();

		/* check, if it was cached */
		if (memcache_lookup() == MEMCACHE_HIT) {
			set(flags, ACL_MEMHIT);
			goto out;
		}

		/* get the acl's now correctly set */
		flags |= setup_acls(check_user(&todo), check_host(&todo),
			check_globals(&todo));
		debug_flags("nach setup_acls():", flags);
		debug_todo("nach setup_acls():", todo);

		/* check for banner */
		if (isset(flags, ACL_FILTER)) {
			flags |= check_banner();
		}
		if (isset(flags, ACL_DENY)) {
			redir_site(REDIR_TYPE_BANNER);
			goto out;
		}
		debug_flags("nach check_banner():", flags);
		debug_todo("nach check_banner():", todo);
		
		/* check for whitelisted url's */
		if (isset(flags, ACL_FILTER))
		if (check_wlist() == ACL_ALLOW) {
			goto out;
		}
		debug_flags("nach check_wlist():", flags);
		debug_todo("nach check_wlist():", todo);

		/* check for blacklisted url's */
		if (isset(flags, ACL_FILTER)) {
			flags |= check_blist();
		}
		if (isset(flags, ACL_DENY)) {
			redir_site(REDIR_TYPE_URL);
			goto out;
		}
		debug_flags("nach check_blist():", flags);
		debug_todo("nach check_blist():", todo);

		/* last step: scan / learn / classify the content */
		if (isset(flags, ACL_FILTER)) {
			check_cont(todo);
		}
		debug_flags("nach check_cont():", flags);
		debug_todo("nach check_cont():", todo);

		out:
		log_flush();

		/* this is currently no real performance improvement, we
		 * need threading for that ... TODO
		 */
		if (isset(flags, ACL_MULTI)) {
			if (!stralloc_cat_uint(&out, squid_nr)) die_nomem();
			if (!stralloc_cats(&out, " ")) die_nomem();
		}

		/* '302:' prefix for methods, other than: HEAD, GET, POST */
		if (isset(flags, ACL_REDIR)) {
			if (redirsite.len)
				if (!stralloc_cats(&out, "302:")) die_nomem();
		}

		/* cache this redirsite */
		if (!isset(flags, ACL_MEMHIT)) {
			memcache_insert();
		}
#if LOGTODO
		else {
			if (isset(logtodo, LOG_MEMCACHE_HITS)) {
				log_headline("squidwall: ");
				log_s("memcache hit on (");
				log_s(squid_url);
				log_s(")\n");
			}
		}
#endif

		if (!stralloc_cat(&out, &redirsite)) die_nomem();
		if (!stralloc_cats(&out, "\n")) die_nomem();
		out.s[out.len]=0;

		/* give squid an empty newline '\n' or the whole redirsite */
		io_wantwrite(1); io_wait(); io_dontwantwrite(1);
		r = write(1, out.s, out.len); /* no perfect errorchecking needed! */
		if (r == -1 && errno == EPIPE) {
			panic("squid has closed the connection!");
			return 111;
		}
#if LOGTODO
		if (isset(logtodo, LOG_SQUID_REQUEST)) {
			log_headline("request");
			if (isset(flags, ACL_MULTI)) {
				log_ulong(squid_nr); log_s(" ");
			}
			log_s(squid_url); log_s(" ");
			log_s(squid_ip); log_s("/");
			log_s(squid_host); log_s(" ");
			log_s(squid_user); log_s(" ");
			log_s(squid_method); log_s("\n");
		}

		if (isset(logtodo, LOG_SQUIDWALL_RESPONSE)) {
			out.s[out.len-1]=0; /* remove \n */
			log_headline("answer");
			log_s(out.s); log_s("\n");
		}
#endif
		log_flush();
	}

	/* never reached */
	return 0;
}
